==> cego-2.6.2.tar.gz: cego/src/CegoOrderSpace.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoOrderSpace.cc
// ------------------
// Cego order space implementation
//     
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2007 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoOrderSpace
// 
// Description: 
//
// Status: CLEAN
//
///////////////////////////////////////////////////////////////////////////////

#include "CegoOrderSpace.h"

#include "Exception.h"
#include "CegoDataType.h"
#include "CegoTableManager.h"
#include "CegoOrderEntry.h"

CegoOrderSpace::CegoOrderSpace(CegoTableManager *pTM)
{
    _pTM = pTM;
    _isCreated = false;
}

CegoOrderSpace::~CegoOrderSpace()
{  
}

void CegoOrderSpace::createOrderSpace(int tabSetId, const Chain& orderSpace, ListT<CegoExpr*>* pOrderList, ListT<Ordering>* pOrderOptList)
{
    _tabSetId = tabSetId;
    _orderSpace = orderSpace;
    _pOrderList = pOrderList;
    _pOrderOptList = pOrderOptList;
    
    CegoExpr **pExpr = pOrderList->First();
    int id=1;
    while ( pExpr )
    {
	ListT<CegoAggregation*> aggList = (*pExpr)->getAggregationList();
	if ( aggList.Size() == 0 )
	{
	    // if no aggregation, we take all used sttributes in order schema

	    ListT<CegoAttrDesc> attrRefList = (*pExpr)->getAttrRefList();
	    
	    CegoAttrDesc* pAttrRef = attrRefList.First();
	    
	    while ( pAttrRef )
	    {
		CegoField f(pAttrRef->getTableName(), pAttrRef->getAttrName());
		f.setId(id);
		_orderSchema.Insert(f);
		id++;
		pAttrRef = attrRefList.Next();
	    }
	}
	else
	{
	    // if aggregation, this is a grouping order, we just take the aggregated values

	    CegoAggregation **pAgg = aggList.First();
	    while ( pAgg )
	    {	    
		CegoField f;
		f.setAttrName("AGG");
		f.setId((*pAgg)->getAggregationId());
		_orderSchema.Insert(f);
		
		pAgg = aggList.Next();
	    }
	}
	pExpr = pOrderList->Next();
    }

    CegoTableObject oe;
    oe = CegoTableObject( _tabSetId, CegoObject::ORDERSPACE, _orderSpace, _orderSchema, orderSpace);
    _pTM->createTableObject(oe);

    _pTM->getObject(_tabSetId, _orderSpace, CegoObject::ORDERSPACE, _oe, _obp);    

    _sysEntry = CegoDataPointer(_obp.getFileId(), _obp.getPageId(), _obp.getEntryPos());

    _pTM->getDBMng()->bufferUnfix(_obp, false, _pTM->getLockHandler());

     _isFirst = true;
     _isCreated = true;
}

void CegoOrderSpace::dropOrderSpace()
{    

    if ( _isCreated )
    {
	if ( _obp.isFixed() )
	    _pTM->getDBMng()->bufferUnfix(_obp, false, _pTM->getLockHandler());
	
	if ( _pTM->objectExists(_tabSetId, _orderSpace, CegoObject::ORDERSPACE) )
	    _pTM->removeObject(_tabSetId, _orderSpace, CegoObject::ORDERSPACE);
    }
    _isCreated = false;
}

void CegoOrderSpace::insertTuple(ListT<CegoField>& addKey, ListT<CegoField>& addTuple)
{

    if ( _isCreated == false )
	throw Exception(EXLOC, "Orderspace not created, cannot insert tuple");

    CegoField* pKF = addKey.First();
    CegoField* pOF = _orderSchema.First();
    while (pKF && pOF )
    {        
        pKF->setId(pOF->getId());
        pKF = addKey.Next();
        pOF = _orderSchema.Next();        
    }

    int id=1;
    CegoField *pF = addTuple.First();
    while (pF)
    {
	pF->setId(id);
	id++;
	pF = addTuple.Next();
    }

    if ( _isFirst )
    {

	_isFirst = false;
	
	// enocode order list and fvl here
	
	char *keyPtr;
	int keyLen;

	CegoTupleState ts = COMMITTED;
	
	_qh.encodeFVL(0, 0, ts, addKey, keyPtr, keyLen);
	
	char *tupPtr;
	int tupLen;
	
	_qh.encodeFVL(0, 0, ts, addTuple, tupPtr, tupLen);
	
	CegoDataPointer nil;
	CegoOrderEntry noe;
	noe.initEntry(keyPtr, keyLen, tupPtr, tupLen, nil, nil, nil);

	try
	{
	    _rdp = _pTM->insertData(_sysEntry, _oe, (char*)noe.getPtr(), noe.getLen(), true);	    
	}
	catch ( Exception e )
	{
	    throw Exception(EXLOC, "Cannot insert order entry", e);
	}

	free(keyPtr);	    
	free(tupPtr);

    }
    else
    {
	insertTupleInTree(addKey, addTuple);
    }
}

CegoOrderCursor* CegoOrderSpace::getCursor()
{
    if ( _isCreated == false )
	throw Exception(EXLOC, "Orderspace not created, get cursor");

    CegoOrderCursor *pOC = new CegoOrderCursor(_tabSetId, _orderSpace, _pTM);
    return pOC;
}


void CegoOrderSpace::insertTupleInTree(ListT<CegoField>& addKey, ListT<CegoField>& addTuple)
{

    ListT<CegoField> actKey = _oe.getSchema();

    char* p;
    int len;

    int tid = 0;
    int tastep = 0;

    CegoDataPointer pdp = _rdp;
    
    // unsigned long lockId;
    CegoBufferPage bp;

    // lockId = _pTM->claimOrderDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
    bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, pdp, p, len);
    
    CegoDataPointer nil;
    CegoDataPointer ndp;

    CegoOrderEntry actEntry;
    actEntry.setPtr(p, len);

    CegoTupleState ts;    
  
    _qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
        
    while ( 1 ) 
    {

	if  ( compKey(actKey, addKey) == HIGHER ) 
	{
	    
	    if (  actEntry.getLeft() == nil )
	    {

		// insert at left
				
		char *keyPtr;
		int keyLen;
		_qh.encodeFVL(tid, tastep, ts, addKey, keyPtr, keyLen);
		
		char *tupPtr;
		int tupLen;
		_qh.encodeFVL(tid, tastep, ts, addTuple, tupPtr, tupLen);
		
		CegoOrderEntry newEntry;
		newEntry.initEntry((char*)keyPtr, keyLen, (char*)tupPtr, tupLen, nil, nil, pdp);

		free(keyPtr);
		free(tupPtr);
				
		CegoDataPointer idp;
		idp = _pTM->insertData(_sysEntry, _oe, (char*)newEntry.getPtr(), newEntry.getLen(), true);

		// cout << "Left IDP=" << idp.toChain() << " ( PDP = " << pdp.toChain() << ")" <<  endl;
		
		actEntry.setLeft(idp);		
		
		// _pTM->releaseOrderDataPtr(lockId, true);
		_pTM->releaseDataPtrUnlocked(bp, true);
		
		return;
		
	    }
	    else
	    {
		pdp = actEntry.getLeft();

		// _pTM->releaseOrderDataPtr(lockId, false);
		_pTM->releaseDataPtrUnlocked(bp, false);

		// lockId = _pTM->claimOrderDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
		bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, pdp, p, len);
		
		actEntry.setPtr(p, len);
		_qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
	    }
	}
	else
	{
	    
	    if ( actEntry.getRight() == nil )
	    {

		// insert at right
		
		char *keyPtr;
		int keyLen;
		_qh.encodeFVL(tid, tastep, ts, addKey, keyPtr, keyLen);
		
		char *tupPtr;
		int tupLen;
		_qh.encodeFVL(tid, tastep, ts, addTuple, tupPtr, tupLen);
		
		CegoOrderEntry newEntry;
		newEntry.initEntry((char*)keyPtr, keyLen, (char*)tupPtr, tupLen, nil, nil, pdp);
		
		free(keyPtr);
		free(tupPtr);
		
		CegoDataPointer idp;
		idp = _pTM->insertData(_sysEntry, _oe, (char*)newEntry.getPtr(), newEntry.getLen(), true);
				
		// cout << "Right IDP=" << idp.toChain() << " ( PDP = " << pdp.toChain() << ")" <<  endl;

		actEntry.setRight(idp);

		// _pTM->releaseOrderDataPtr(lockId, true);
		_pTM->releaseDataPtrUnlocked(bp, true);

		return;
		
	    }
	    else
	    {
		pdp = actEntry.getRight();

		// _pTM->releaseOrderDataPtr(lockId, false);
		_pTM->releaseDataPtrUnlocked(bp, false);		

		// lockId = _pTM->claimOrderDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
		bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::SYNC, pdp, p, len);

		actEntry.setPtr(p, len);
		_qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
	    }	    
	}	
    }
}

CegoOrderSpace::KeyComp CegoOrderSpace::compKey(ListT<CegoField>& key1, ListT<CegoField>& key2)
{

    CegoExpr **pExpr = _pOrderList->First();
    Ordering *pOrd = _pOrderOptList->First();
   
    while ( pExpr && pOrd )
    {

	setAggregationValue(*pExpr, key1);	
	(*pExpr)->setFieldListArray(&key1, 1);
	CegoFieldValue fv1 = (*pExpr)->evalFieldValue();

	setAggregationValue(*pExpr, key2);
	(*pExpr)->setFieldListArray(&key2, 1);
	CegoFieldValue fv2 = (*pExpr)->evalFieldValue();

	if ( fv1 < fv2 )
	{
	    if ( *pOrd == ASC )
		return LOWER;
	    else
		return HIGHER;
	}
	else if ( fv1 > fv2 )
	{
	    if ( *pOrd == ASC )
		return HIGHER;
	    else
		return LOWER;
	}

	pExpr = _pOrderList->Next();
	pOrd = _pOrderOptList->Next();
    }

    return EQUAL;
}

void CegoOrderSpace::setAggregationValue(CegoExpr *pExpr, ListT<CegoField>& fl)
{
    ListT<CegoAggregation*> aggList = pExpr->getAggregationList();
    
    CegoAggregation **pAgg = aggList.First();
    while ( pAgg )
    {
	CegoField *pF = fl.First();
	while ( pF )
	{
	    if ( pF->getId() == (*pAgg)->getAggregationId() )
	    {	
		(*pAgg)->setFieldValue(pF->getValue());
		pF = 0;
	    }
	    else
	    {
		pF = fl.Next();
	    }
	}
	pAgg = aggList.Next();
    }
}

==> cego-2.6.8.tar.gz: cego/src/CegoOrderSpace.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoOrderSpace.cc
// ------------------
// Cego order space implementation
//     
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoOrderSpace
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

// base includes
#include "Exception.h"

// cego includes
#include "CegoOrderSpace.h"
#include "CegoDataType.h"
#include "CegoTableManager.h"
#include "CegoOrderEntry.h"

CegoOrderSpace::CegoOrderSpace(CegoTableManager *pTM)
{
    _pTM = pTM;
    _isCreated = false;
}

CegoOrderSpace::~CegoOrderSpace()
{  
}

void CegoOrderSpace::createOrderSpace(int tabSetId, const Chain& orderSpace, ListT<CegoExpr*>* pOrderList, ListT<Ordering>* pOrderOptList)
{
    _tabSetId = tabSetId;
    _orderSpace = orderSpace;
    _pOrderList = pOrderList;
    _pOrderOptList = pOrderOptList;
    
    CegoExpr **pExpr = pOrderList->First();
    int id=1;
    while ( pExpr )
    {
	ListT<CegoAggregation*> aggList = (*pExpr)->getAggregationList();
	if ( aggList.Size() == 0 )
	{
	    // if no aggregation, we take all used sttributes in order schema

	    ListT<CegoAttrDesc> attrRefList = (*pExpr)->getAttrRefList();
	    
	    CegoAttrDesc* pAttrRef = attrRefList.First();
	    
	    while ( pAttrRef )
	    {
		CegoField f(pAttrRef->getTableName(), pAttrRef->getAttrName());
		f.setId(id);
		_orderSchema.Insert(f);
		id++;
		pAttrRef = attrRefList.Next();
	    }
	}
	else
	{
	    // if aggregation, this is a grouping order, we just take the aggregated values

	    CegoAggregation **pAgg = aggList.First();
	    while ( pAgg )
	    {	    
		CegoField f;
		f.setAttrName("AGG");
		f.setId((*pAgg)->getAggregationId());
		_orderSchema.Insert(f);
		
		pAgg = aggList.Next();
	    }
	}
	pExpr = pOrderList->Next();
    }

    CegoTableObject oe;
    oe = CegoTableObject( _tabSetId, CegoObject::ORDERSPACE, _orderSpace, _orderSchema, orderSpace);
    _pTM->createTableObject(oe);

    _pTM->getObject(_tabSetId, _orderSpace, CegoObject::ORDERSPACE, _oe, _obp);    

    _sysEntry = CegoDataPointer(_obp.getFileId(), _obp.getPageId(), _obp.getEntryPos());

    // cout << "OrderSpace SE = " << _sysEntry << endl;

    // int semid = _sysEntry.getFileId() * _sysEntry.getPageId() + _sysEntry.getOffset();
    // cout << "SemId = " << semid << endl;

    _pTM->getDBMng()->bufferUnfix(_obp, false, _pTM->getLockHandler());

     _isFirst = true;
     _isCreated = true;
}

void CegoOrderSpace::dropOrderSpace()
{    

    if ( _isCreated )
    {
	if ( _obp.isFixed() )
	    _pTM->getDBMng()->bufferUnfix(_obp, false, _pTM->getLockHandler());
	
	if ( _pTM->objectExists(_tabSetId, _orderSpace, CegoObject::ORDERSPACE) )
	    _pTM->removeObject(_tabSetId, _orderSpace, CegoObject::ORDERSPACE);
    }
    _isCreated = false;
}

void CegoOrderSpace::insertTuple(ListT<CegoField>& addKey, ListT<CegoField>& addTuple)
{

    if ( _isCreated == false )
	throw Exception(EXLOC, "Orderspace not created, cannot insert tuple");

    CegoField* pKF = addKey.First();
    CegoField* pOF = _orderSchema.First();
    while (pKF && pOF )
    {        
        pKF->setId(pOF->getId());
        pKF = addKey.Next();
        pOF = _orderSchema.Next();        
    }

    int id=1;
    CegoField *pF = addTuple.First();
    while (pF)
    {
	pF->setId(id);
	id++;
	pF = addTuple.Next();
    }

    if ( _isFirst )
    {

	_isFirst = false;
	
	// enocode order list and fvl here
	
	char *keyPtr;
	int keyLen;

	CegoTupleState ts = COMMITTED;
	
	_qh.encodeFVL(0, 0, ts, addKey, keyPtr, keyLen);
	
	char *tupPtr;
	int tupLen;
	
	_qh.encodeFVL(0, 0, ts, addTuple, tupPtr, tupLen);
	
	CegoDataPointer nil;
	CegoOrderEntry noe;
	noe.initEntry(keyPtr, keyLen, tupPtr, tupLen, nil, nil, nil);

	try
	{
	    _rdp = _pTM->insertData(_sysEntry, _oe, (char*)noe.getPtr(), noe.getLen(), true, true);	    
	}
	catch ( Exception e )
	{
	    throw Exception(EXLOC, "Cannot insert order entry", e);
	}

	free(keyPtr);	    
	free(tupPtr);

    }
    else
    {
	insertTupleInTree(addKey, addTuple);
    }
}

CegoOrderCursor* CegoOrderSpace::getCursor()
{
    if ( _isCreated == false )
	throw Exception(EXLOC, "Orderspace not created, get cursor");

    CegoOrderCursor *pOC = new CegoOrderCursor(_tabSetId, _orderSpace, _pTM);
    return pOC;
}

void CegoOrderSpace::insertTupleInTree(ListT<CegoField>& addKey, ListT<CegoField>& addTuple)
{

    ListT<CegoField> actKey = _oe.getSchema();

    char* p;
    int len;

    int tid = 0;
    int tastep = 0;

    CegoDataPointer pdp = _rdp;
    
    // unsigned long lockId;
    CegoBufferPage bp;

    // lockId = _pTM->claimOrderDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
    bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::NOSYNC, pdp, p, len);
    
    CegoDataPointer nil;
    CegoDataPointer ndp;

    CegoOrderEntry actEntry;
    actEntry.setPtr(p, len);

    CegoTupleState ts;    
  
    _qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
        
    while ( 1 ) 
    {

	if  ( compKey(actKey, addKey) == HIGHER ) 
	{
	    
	    if (  actEntry.getLeft() == nil )
	    {

		// insert at left
				
		char *keyPtr;
		int keyLen;
		_qh.encodeFVL(tid, tastep, ts, addKey, keyPtr, keyLen);
		
		char *tupPtr;
		int tupLen;
		_qh.encodeFVL(tid, tastep, ts, addTuple, tupPtr, tupLen);
		
		CegoOrderEntry newEntry;
		newEntry.initEntry((char*)keyPtr, keyLen, (char*)tupPtr, tupLen, nil, nil, pdp);

		free(keyPtr);
		free(tupPtr);
				
		CegoDataPointer idp;
		idp = _pTM->insertData(_sysEntry, _oe, (char*)newEntry.getPtr(), newEntry.getLen(), true, true);

		// cout << "Left IDP=" << idp.toChain() << " ( PDP = " << pdp.toChain() << ")" <<  endl;
		
		actEntry.setLeft(idp);		
		
		// _pTM->releaseOrderDataPtr(lockId, true);
		_pTM->releaseDataPtrUnlocked(bp, true);
		
		return;
		
	    }
	    else
	    {
		pdp = actEntry.getLeft();

		// _pTM->releaseOrderDataPtr(lockId, false);
		_pTM->releaseDataPtrUnlocked(bp, false);

		// lockId = _pTM->claimOrderDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
		bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::NOSYNC, pdp, p, len);
		
		actEntry.setPtr(p, len);
		_qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
	    }
	}
	else
	{
	    
	    if ( actEntry.getRight() == nil )
	    {

		// insert at right
		
		char *keyPtr;
		int keyLen;
		_qh.encodeFVL(tid, tastep, ts, addKey, keyPtr, keyLen);
		
		char *tupPtr;
		int tupLen;
		_qh.encodeFVL(tid, tastep, ts, addTuple, tupPtr, tupLen);
		
		CegoOrderEntry newEntry;
		newEntry.initEntry((char*)keyPtr, keyLen, (char*)tupPtr, tupLen, nil, nil, pdp);
		
		free(keyPtr);
		free(tupPtr);
		
		CegoDataPointer idp;
		idp = _pTM->insertData(_sysEntry, _oe, (char*)newEntry.getPtr(), newEntry.getLen(), true, true);
				
		// cout << "Right IDP=" << idp.toChain() << " ( PDP = " << pdp.toChain() << ")" <<  endl;

		actEntry.setRight(idp);

		// _pTM->releaseOrderDataPtr(lockId, true);
		_pTM->releaseDataPtrUnlocked(bp, true);

		return;
		
	    }
	    else
	    {
		pdp = actEntry.getRight();

		// _pTM->releaseOrderDataPtr(lockId, false);
		_pTM->releaseDataPtrUnlocked(bp, false);		

		// lockId = _pTM->claimOrderDataPtr(_tabSetId, CegoLockHandler::WRITE, CegoBufferPool::SYNC, pdp, p, len);
		bp = _pTM->claimDataPtrUnlocked(_tabSetId, CegoBufferPool::NOSYNC, pdp, p, len);

		actEntry.setPtr(p, len);
		_qh.decodeFVL(actKey, (char*)actEntry.getKeyPtr(), actEntry.getKeyLen(), tid, tastep, ts);
	    }	    
	}	
    }
}

CegoOrderSpace::KeyComp CegoOrderSpace::compKey(ListT<CegoField>& key1, ListT<CegoField>& key2)
{

    CegoExpr **pExpr = _pOrderList->First();
    Ordering *pOrd = _pOrderOptList->First();
   
    while ( pExpr && pOrd )
    {

	setAggregationValue(*pExpr, key1);	
	(*pExpr)->setFieldListArray(&key1, 1);
	CegoFieldValue fv1 = (*pExpr)->evalFieldValue();

	setAggregationValue(*pExpr, key2);
	(*pExpr)->setFieldListArray(&key2, 1);
	CegoFieldValue fv2 = (*pExpr)->evalFieldValue();

	if ( fv1 < fv2 )
	{
	    if ( *pOrd == ASC )
		return LOWER;
	    else
		return HIGHER;
	}
	else if ( fv1 > fv2 )
	{
	    if ( *pOrd == ASC )
		return HIGHER;
	    else
		return LOWER;
	}

	pExpr = _pOrderList->Next();
	pOrd = _pOrderOptList->Next();
    }

    return EQUAL;
}

void CegoOrderSpace::setAggregationValue(CegoExpr *pExpr, ListT<CegoField>& fl)
{
    ListT<CegoAggregation*> aggList = pExpr->getAggregationList();
    
    CegoAggregation **pAgg = aggList.First();
    while ( pAgg )
    {
	CegoField *pF = fl.First();
	while ( pF )
	{
	    if ( pF->getId() == (*pAgg)->getAggregationId() )
	    {	
		(*pAgg)->setFieldValue(pF->getValue());
		pF = 0;
	    }
	    else
	    {
		pF = fl.Next();
	    }
	}
	pAgg = aggList.Next();
    }
}
